@@ -1,6 +1,6 @@
 from collections import namedtuple
+from enum import IntEnum
 from typing import Dict
-import distutils.ccompiler
 import casadi as ca
 import numpy as np
 import copy
@@ -11,14 +11,21 @@ import logging
 import pickle
 import contextlib
 
-from pymoca import parser, tree, ast, __version__
+from pymoca import __version__
 from . import generator
 from .alias_relation import AliasRelation
 from .model import CASADI_ATTRIBUTES, Model, Variable, DelayArgument
+from ._options import _merge_default_options, _get_default_options as get_default_options
 
 logger = logging.getLogger("pymoca")
 
 
+class _DepMeta(IntEnum):
+    NOT_MX = 0
+    MX_DEPENDENT = 1
+    MX_INDEPENDENT = 2
+
+
 class CachedModel(Model):
 
     def __init__(self):
@@ -84,9 +91,14 @@ class InvalidCacheError(Exception):
     pass
 
 def _compile_model(model_folder: str, model_name: str, compiler_options: Dict[str, str]):
+    # Importing the antlr4 (generated modules) is rather slow. Avoid for this
+    # ~100 ms startup overhead for cached models by importing only when
+    # compiling.
+    from pymoca import parser
+
     # Load folders
     tree = None
-    for folder in [model_folder] + compiler_options.get('library_folders', []):
+    for folder in [model_folder] + compiler_options['library_folders']:
         for root, dir, files in os.walk(folder, followlinks=True):
             for item in fnmatch.filter(files, "*.mo"):
                 logger.info("Parsing {}".format(item))
@@ -101,12 +113,12 @@ def _compile_model(model_folder: str, model_name: str, compiler_options: Dict[st
     logger.info("Generating CasADi model")
 
     model = generator.generate(tree, model_name, compiler_options)
-    if compiler_options.get('check_balanced', True):
+    if compiler_options['check_balanced']:
         model.check_balanced()
 
     model.simplify(compiler_options)
 
-    if compiler_options.get('verbose', False):
+    if compiler_options['verbose']:
         model.check_balanced()
 
     model._post_checks()
@@ -114,6 +126,10 @@ def _compile_model(model_folder: str, model_name: str, compiler_options: Dict[st
     return model
 
 def _codegen_model(model_folder: str, f: ca.Function, library_name: str):
+    # Avoid locating compiler when loading from cache by loading
+    # distutils.ccompiler here on-demand.
+    import distutils.ccompiler
+
     # Compile shared libraries
     if os.name == 'posix':
         compiler_flags = ['-O2', '-fPIC']
@@ -167,11 +183,13 @@ def save_model(model_folder: str, model_name: str, model: Model,
     :param compiler_options: Dictionary of compiler options.
     """
 
+    compiler_options = _merge_default_options(compiler_options)
+
     objects = {'dae_residual': None, 'initial_residual': None, 'variable_metadata': None, 'delay_arguments': None}
     for o in objects.keys():
         f = getattr(model, o + '_function')
 
-        if compiler_options.get('codegen', False):
+        if compiler_options['codegen']:
             objects[o] = _codegen_model(model_folder, f, '{}_{}'.format(model_name, o))
         else:
             objects[o] = f
@@ -207,13 +225,18 @@ def save_model(model_folder: str, model_name: str, model: Model,
 
         for k, key in enumerate(['states', 'alg_states', 'inputs', 'parameters', 'constants']):
             metadata_shape = (len(getattr(model, key)), len(CASADI_ATTRIBUTES))
-            m = db[key + "__metadata_dependent"] = np.zeros(metadata_shape, dtype=bool)
+            m = db[key + "__metadata_dependent"] = np.zeros(metadata_shape, dtype=int)
+            if np.prod(m.shape) > 0:
+                assert m[0, 0] == _DepMeta.NOT_MX
+
             for i, v in enumerate(getattr(model, key)):
                 for j, tmp in enumerate(CASADI_ATTRIBUTES):
                     attr = getattr(v, tmp)
-                    if (isinstance(attr, ca.MX) and not attr.is_constant()
-                        and ca.depends_on(attr, parameter_vector)):
-                        m[i, j] = True
+                    if isinstance(attr, ca.MX):
+                        if not attr.is_constant() and ca.depends_on(attr, parameter_vector):
+                            m[i, j] = _DepMeta.MX_DEPENDENT
+                        else:
+                            m[i, j] = _DepMeta.MX_INDEPENDENT
 
         # Delay dependency checking
         if model.delay_states:
@@ -254,12 +277,14 @@ def load_model(model_folder: str, model_name: str, compiler_options: Dict[str, s
     :returns: CachedModel instance.
     """
 
+    compiler_options = _merge_default_options(compiler_options)
+
     db_file = os.path.join(model_folder, model_name + ".pymoca_cache")
 
-    if compiler_options.get('mtime_check', True):
+    if compiler_options['mtime_check']:
         # Mtime check
         cache_mtime = os.path.getmtime(db_file)
-        for folder in [model_folder] + compiler_options.get('library_folders', []):
+        for folder in [model_folder] + compiler_options['library_folders']:
             for root, dir, files in os.walk(folder, followlinks=True):
                 for item in fnmatch.filter(files, "*.mo"):
                     filename = os.path.join(root, item)
@@ -271,7 +296,13 @@ def load_model(model_folder: str, model_name: str, compiler_options: Dict[str, s
 
     # Load metadata
     with open(db_file, 'rb') as f:
-        db = pickle.load(f)
+        try:
+            db = pickle.load(f)
+        except RuntimeError as e:
+            if "DeserializingStream" in str(e):
+                raise InvalidCacheError('Cache generated for incompatible CasADi version')
+            else:
+                raise
 
         if db['version'] != __version__:
             raise InvalidCacheError('Cache generated for a different version of pymoca')
@@ -287,7 +318,7 @@ def load_model(model_folder: str, model_name: str, compiler_options: Dict[str, s
             raise InvalidCacheError('Cache generated for different compiler options')
 
         # Pickles are platform independent, but dynamic libraries are not
-        if compiler_options.get('codegen', False):
+        if compiler_options['codegen']:
             if db['library_os'] != os.name:
                 raise InvalidCacheError('Cache generated for incompatible OS')
 
@@ -323,17 +354,23 @@ def load_model(model_folder: str, model_name: str, compiler_options: Dict[str, s
         metadata = dict(zip(variables_with_metadata, model.variable_metadata_function(parameter_vector)))
         independent_metadata = dict(zip(
             variables_with_metadata,
-            (np.array(x) for x in model.variable_metadata_function(ca.veccat(*[np.nan for v in model.parameters])))))
+            (ca.MX(x) for x in model.variable_metadata_function(ca.veccat(*[np.nan for v in model.parameters])))))
 
         for k, key in enumerate(variables_with_metadata):
             m = db[key + "__metadata_dependent"]
             for i, d in enumerate(db[key]):
                 variable = variable_dict[d['name']]
                 for j, tmp in enumerate(CASADI_ATTRIBUTES):
-                    if m[i, j]:
+                    if m[i, j] == _DepMeta.MX_DEPENDENT:
                         setattr(variable, tmp, metadata[key][i, j])
-                    else:
+                    elif m[i, j] == _DepMeta.MX_INDEPENDENT:
                         setattr(variable, tmp, independent_metadata[key][i, j])
+                    else:
+                        # Already handled as part of Variable dict. That way
+                        # we also do not have to worry about making sure the
+                        # type of the cached model is exactly the same, as
+                        # pickling ensures that.
+                        pass
 
         # Evaluate delay arguments:
         if model.delay_states:
@@ -366,43 +403,53 @@ def load_model(model_folder: str, model_name: str, compiler_options: Dict[str, s
 
             duration_dependencies = db['__delay_duration_dependent']
 
+            # Get rid of false dependency symbols not used in any delay
+            # durations. This significantly reduces the work the (slow)
+            # substitute() calls have to do later on.
+            actual_deps = sorted(set(np.array(duration_dependencies).ravel()))
+
+            actual_dep_symbols = [np.nan] * len(all_symbols)
+            for i in actual_deps:
+                actual_dep_symbols[i] = all_symbols[i]
+
+            delay_durations_simplified = ca.Function(
+                'replace_false_deps',
+                all_symbols,
+                delay_durations_raw).call(
+                    actual_dep_symbols)
+
+            # Get rid of remaining hidden dependencies in the delay durations
             for i, expr in enumerate(delay_expressions_raw):
                 if duration_dependencies[i]:
-                    dur = delay_durations_raw[i]
+                    dur = delay_durations_simplified[i]
 
-                    deps = set(ca.symvar(dur))
-                    actual_deps = {all_symbols[j] for j in duration_dependencies[i]}
-                    false_deps = deps - actual_deps
+                    if len(duration_dependencies[i]) < len(actual_deps):
+                        deps = set(ca.symvar(dur))
+                        actual_deps = {all_symbols[j] for j in duration_dependencies[i]}
+                        false_deps = deps - actual_deps
 
-                    if false_deps:
-                        [dur] = ca.substitute(
-                            [dur],
-                            list(false_deps),
-                            [ca.repmat(np.nan, *d.size()) for d in false_deps])
+                        if false_deps:
+                            [dur] = ca.substitute(
+                                [dur],
+                                list(false_deps),
+                                [np.nan] * len(false_deps))
+                    else:
+                        # Already removed all false dependencies
+                        pass
                 else:
                     dur = independent_delay_durations_raw[i]
 
                 model.delay_arguments.append(DelayArgument(expr, dur))
 
-        # Try to coerce parameters into their Python types
-        for p in model.parameters:
-            for attr in CASADI_ATTRIBUTES:
-                v = getattr(p, attr)
-                v_mx = ca.MX(v)
-                if v_mx.is_constant() and v_mx.is_regular():
-                    setattr(p, attr, p.python_type(v))
-
     # Done
     return model
 
 def transfer_model(model_folder: str, model_name: str, compiler_options=None):
-    if compiler_options is None:
-        compiler_options = {}
-    else:
-        compiler_options = copy.copy(compiler_options)
 
-    cache = compiler_options.setdefault('cache', False)
-    codegen = compiler_options.setdefault('codegen', False)
+    compiler_options = _merge_default_options(compiler_options)
+
+    cache = compiler_options['cache']
+    codegen = compiler_options['codegen']
 
     # Compilation takes precedence over caching, and disables it
     if cache and codegen:
@@ -414,7 +461,7 @@ def transfer_model(model_folder: str, model_name: str, compiler_options=None):
         # expanding to SX. We only raise a warning when we have to (re)compile
         # the model though.
         raise_expand_warning = False
-        if cache and not compiler_options.get('expand_mx', False):
+        if cache and not compiler_options['expand_mx']:
             compiler_options['expand_mx'] = True
             raise_expand_warning = True
 
@@ -428,3 +475,5 @@ def transfer_model(model_folder: str, model_name: str, compiler_options=None):
             return model
     else:
         return _compile_model(model_folder, model_name, compiler_options)
+
+
