@@ -4,17 +4,26 @@ import numpy as np
 import itertools
 import logging
 import re
+import sys
 
 from pymoca import ast
+
 from .alias_relation import AliasRelation
 from .mtensor import _MTensor
+from ._options import _merge_default_options
 
 logger = logging.getLogger("pymoca")
 
-CASADI_COMPARISON_DEPTH = 100
+CASADI_COMPARISON_DEPTH = sys.maxsize
+SUBSTITUTE_LOOP_LIMIT = 100
+SIMPLIFICATION_LOOP_LIMIT = 50
 CASADI_ATTRIBUTES = [attr for attr in ast.Symbol.ATTRIBUTES if not attr == 'unit']
 
 
+class _DefaultValue(int):
+    pass
+
+
 class Variable:
     def __init__(self, symbol, python_type=float, aliases=None):
         if aliases is None:
@@ -25,10 +34,10 @@ class Variable:
 
         # Default attribute values
         self.value = np.nan
-        self.start = 0
+        self.start = _DefaultValue(0)
         self.min = -np.inf
         self.max = np.inf
-        self.nominal = 1
+        self.nominal = 0
         self.fixed = False
 
     def __str__(self):
@@ -43,12 +52,22 @@ class Variable:
         d['shape'] = (self.symbol.size1(), self.symbol.size2())
         d['python_type'] = self.python_type
         d['aliases'] = self.aliases
+
+        for attr in CASADI_ATTRIBUTES:
+            d[attr] = getattr(self, attr)
+            if isinstance(d[attr], ca.MX):
+                d[attr] = None  # Will be filled using variable_metadata_function
+
         return d
 
     @classmethod
     def from_dict(cls, d):
         variable = cls(ca.MX.sym(d['name'], *d['shape']), d['python_type'])
         variable.aliases = d['aliases']
+
+        for attr in CASADI_ATTRIBUTES:
+            setattr(variable, attr, d[attr])
+
         return variable
 
 
@@ -174,6 +193,23 @@ class Model:
 
         return equations
 
+    def _substitute_metadata(self, symbols, values):
+        substitutions = []
+        for variable in itertools.chain(self.states, self.alg_states, self.inputs, self.parameters, self.constants):
+            for attribute in CASADI_ATTRIBUTES:
+                value = getattr(variable, attribute)
+                if isinstance(value, ca.MX) and not value.is_constant():
+                    substitutions.append((value, variable, attribute))
+
+        if len(substitutions) == 0:
+            return
+
+        expressions, variables, attributes = zip(*substitutions)
+        expressions = ca.substitute(expressions, symbols, values)
+
+        for variable, attribute, value in zip(variables, attributes, expressions):
+            setattr(variable, attribute, value)
+
     def _expand_vectors(self):
         logger.info("Expanding vectors")
 
@@ -207,7 +243,16 @@ class Model:
                         # a was a scalar, b a 1d array of size [3] and c a 2d array of size
                         # [1,3].
                         modelica_shape = old_var.symbol._modelica_shape
-                        symbol_names = old_var.symbol.name().split('.')
+                        old_sym_name = old_var.symbol.name()
+
+                        # Expanded derivative symbols get their indices inside the parantheses.
+                        # For a symbol called "der(a.x)", the prefix will be "der(", the postfix
+                        # will be ")", with old_sym_name becoming "a.x".
+                        # For a symbol called "a.x", the prefix and postfix will be empty strings.
+                        prefix, old_sym_name, postfix = \
+                            re.match(r'((?:der\()*|\b)(.*?)([\)]*|\b)$', old_sym_name).groups()
+
+                        symbol_names = old_sym_name.split('.')
                         assert len(symbol_names) == len(modelica_shape)
                         component_name_format = ''
                         for i, symbol_name in enumerate(symbol_names):
@@ -220,6 +265,8 @@ class Model:
                                     ','.join('{}' for _ in range(len(modelica_shape[i]))) + \
                                     ']'
 
+                        component_name_format = prefix + component_name_format + postfix
+
                         iterator_shape = tuple([d for var_shape in modelica_shape
                                                 for d in var_shape if d is not None])
 
@@ -231,15 +278,29 @@ class Model:
                         for attribute in CASADI_ATTRIBUTES:
                             # Can't convert 3D arrays to MX, so we convert to nparray instead
                             value = getattr(old_var, attribute)
-                            if not isinstance(value, ca.MX) and not np.isscalar(value):
-                                value = np.array(value)
+                            if not isinstance(value, ca.MX):
+                                if np.isscalar(value):
+                                    # Just assign without indexing
+                                    val = value
+                                elif isinstance(value, list):
+                                    val = value
+                                    for i in ind:
+                                        val = val[i]
+                                elif isinstance(value, (ca.DM, np.ndarray)):
+                                    val = value[ind]
+                                    if old_var.python_type in {float, int}:
+                                        val = old_var.python_type(val)
+                                else:
+                                    assert False, "Unexpected type from CasADi generator"
                             else:
-                                value = ca.MX(getattr(old_var, attribute))
+                                if np.prod(value.shape) == 1:
+                                    # Just assign without indexing
+                                    val = value
+                                else:
+                                    val = value[ind]
+
+                            setattr(component_var, attribute, val)
 
-                            if np.prod(value.shape) == 1:
-                                setattr(component_var, attribute, value)
-                            else:
-                                setattr(component_var, attribute, value[ind])
                         expanded_symbols.append(component_var)
 
                     s = old_var.symbol._mx if isinstance(old_var.symbol, _MTensor) else old_var.symbol
@@ -293,15 +354,31 @@ class Model:
         assert set(self.delay_states).issubset(input_names)
 
         # Replace values in metadata
-        for variable in itertools.chain(self.states, self.alg_states, self.inputs, self.parameters, self.constants):
-            for attribute in CASADI_ATTRIBUTES:
-                value = getattr(variable, attribute)
-                if isinstance(value, ca.MX) and not value.is_constant():
-                    [value] = ca.substitute([value], symbols, values)
-                    setattr(variable, attribute, value)
+        self._substitute_metadata(symbols, values)
 
     def simplify(self, options):
-        if options.get('expand_vectors', False) and options.get('expand_mx', False):
+        options = _merge_default_options(options)
+
+        alg_states_left = 0
+
+        for simplification_iter in range(SIMPLIFICATION_LOOP_LIMIT):
+            if options.get('iterative_simplification', False):
+                logger.info("Simplification iteration {}".format(simplification_iter + 1))
+
+            self._simplify_once(options)
+
+            if options.get('iterative_simplification', False) and alg_states_left != len(self.alg_states):
+                alg_states_left = len(self.alg_states)
+                simplification_iter += 1
+            else:
+                break
+        else:
+            logger.warning("Simplification exceeded maximum iteration limit.")
+
+        logger.info("Finished model simplification")
+
+    def _simplify_once(self, options):
+        if options['expand_vectors'] and options['expand_mx']:
             # If we are _not_ expanding MX to SX, we do the expansion of
             # vectors first, such that we are be able to detect more aliases
             # (e.g individual array elements, or elements in a for loop).
@@ -313,7 +390,7 @@ class Model:
             self.equations = self._expand_simplify_mx(self.equations)
             self.initial_equations = self._expand_simplify_mx(self.initial_equations)
 
-        if options.get('replace_parameter_expressions', False):
+        if options['replace_parameter_expressions']:
             logger.info("Replacing parameter expressions")
 
             simple_parameters, symbols, values = [], [], []
@@ -339,11 +416,14 @@ class Model:
             if len(values) > 0:
                 # Resolve expressions that include other, non-simple parameter
                 # expressions.
-                converged = False
-                while not converged:
+                for i in range(SUBSTITUTE_LOOP_LIMIT):
                     new_values = ca.substitute(values, symbols, values)
                     converged = ca.is_equal(ca.veccat(*values), ca.veccat(*new_values), CASADI_COMPARISON_DEPTH)
                     values = new_values
+                    if converged:
+                        break
+                else:
+                    logger.warning("Substitution of expressions exceeded maximum iteration limit.")
 
                 if len(self.equations) > 0:
                     self.equations = ca.substitute(self.equations, symbols, values)
@@ -353,14 +433,9 @@ class Model:
                     self.delay_arguments = self._substitute_delay_arguments(self.delay_arguments, symbols, values)
 
                 # Replace parameter expressions in metadata
-                for variable in itertools.chain(self.states, self.alg_states, self.inputs, self.parameters, self.constants):
-                    for attribute in CASADI_ATTRIBUTES:
-                        value = getattr(variable, attribute)
-                        if isinstance(value, ca.MX) and not value.is_constant():
-                            [value] = ca.substitute([value], symbols, values)
-                            setattr(variable, attribute, value)
+                self._substitute_metadata(symbols, values)
 
-        if options.get('replace_constant_expressions', False):
+        if options['replace_constant_expressions']:
             logger.info("Replacing constant expressions")
 
             simple_constants, symbols, values = [], [], []
@@ -377,11 +452,14 @@ class Model:
             if len(values) > 0:
                 # Resolve expressions that include other, non-simple parameter
                 # expressions.
-                converged = False
-                while not converged:
+                for i in range(SUBSTITUTE_LOOP_LIMIT):
                     new_values = ca.substitute(values, symbols, values)
                     converged = ca.is_equal(ca.veccat(*values), ca.veccat(*new_values), CASADI_COMPARISON_DEPTH)
                     values = new_values
+                    if converged:
+                        break
+                else:
+                    logger.warning("Substitution of expressions exceeded maximum iteration limit.")
 
                 if len(self.equations) > 0:
                     self.equations = ca.substitute(self.equations, symbols, values)
@@ -391,14 +469,9 @@ class Model:
                     self.delay_arguments = self._substitute_delay_arguments(self.delay_arguments, symbols, values)
 
                 # Replace constant expressions in metadata
-                for variable in itertools.chain(self.states, self.alg_states, self.inputs, self.parameters, self.constants):
-                    for attribute in CASADI_ATTRIBUTES:
-                        value = getattr(variable, attribute)
-                        if isinstance(value, ca.MX) and not value.is_constant():
-                            [value] = ca.substitute([value], symbols, values)
-                            setattr(variable, attribute, value)
+                self._substitute_metadata(symbols, values)
 
-        if options.get('eliminate_constant_assignments', False):
+        if options['eliminate_constant_assignments']:
             logger.info("Elimating constant variable assignments")
 
             alg_states = OrderedDict([(s.symbol.name(), s) for s in self.alg_states])
@@ -445,7 +518,7 @@ class Model:
             self.alg_states = list(alg_states.values())
             self.equations = reduced_equations
 
-        if options.get('replace_parameter_values', False):
+        if options['replace_parameter_values']:
             logger.info("Replacing parameter values")
 
             # N.B. Any parameter expression elimination must be done first.
@@ -464,14 +537,9 @@ class Model:
             self.parameters = unspecified_parameters
 
             # Replace parameter values in metadata
-            for variable in itertools.chain(self.states, self.alg_states, self.inputs, self.parameters, self.constants):
-                for attribute in CASADI_ATTRIBUTES:
-                    value = getattr(variable, attribute)
-                    if isinstance(value, ca.MX) and not value.is_constant():
-                        [value] = ca.substitute([value], symbols, values)
-                        setattr(variable, attribute, value)
+            self._substitute_metadata(symbols, values)
 
-        if options.get('replace_constant_values', False):
+        if options['replace_constant_values']:
             logger.info("Replacing constant values")
 
             # N.B. Any parameter expression elimination must be done first.
@@ -483,77 +551,180 @@ class Model:
                 self.initial_equations = ca.substitute(self.initial_equations, symbols, values)
             if len(self.delay_arguments) > 0:
                 self.delay_arguments = self._substitute_delay_arguments(self.delay_arguments, symbols, values)
+
+            # Also remove the aliases of any of the constants from the alias relation
+            for constant in self.constants:
+                if constant.aliases:
+                    # Note that we will only get here if simplify is called multiple times on the
+                    # same model, as alias detection is done _after_ replacing constant values.
+                    # Otherwise a variable will not have any aliases yet.
+                    self.alias_relation.remove(constant.symbol.name())
+
             self.constants = []
 
             # Replace constant values in metadata
-            for variable in itertools.chain(self.states, self.alg_states, self.inputs, self.parameters, self.constants):
-                for attribute in CASADI_ATTRIBUTES:
-                    value = getattr(variable, attribute)
-                    if isinstance(value, ca.MX) and not value.is_constant():
-                        [value] = ca.substitute([value], symbols, values)
-                        setattr(variable, attribute, value)
+            self._substitute_metadata(symbols, values)
 
-        if options.get('eliminable_variable_expression', None) is not None:
+        if options['eliminable_variable_expression'] is not None:
             logger.info("Elimating variables that match the regular expression {}".format(options['eliminable_variable_expression']))
 
+            # Due to CasADi's ca.is_equal not properly matching short-
+            # circuited MX if-else expressions, we can end up in an infinite
+            # loop. With expanded (to SX and back) equations we are safe, as
+            # SX does not support short-circuiting.
+            if not options['expand_mx']:
+                raise Exception("The use of `eliminable_variable_expression` requires `expand_mx` set to True")
+
             p = re.compile(options['eliminable_variable_expression'])
 
+            states = OrderedDict([(s.symbol.name(), s) for s in self.states])
             alg_states = OrderedDict([(s.symbol.name(), s) for s in self.alg_states])
 
-            variables = []
-            values = []
+            all_states = OrderedDict([(s.symbol.name(), s) for s in itertools.chain(self.states, self.alg_states)])
 
-            reduced_equations = []
-            for eq in self.equations:
-                if eq.is_symbolic() and eq.name() in alg_states and p.match(eq.name()):
-                    variables.append(eq)
-                    values.append(0.0)
-                    del alg_states[eq.name()]
-                    # Skip this equation
-                    continue
+            # NOTE: dictionary keys are the correspondig _state_'s name, not
+            # the names of the derivative states themselves.
+            der_states = OrderedDict(zip(states.keys(), self.der_states))
+
+            def extract_assignment(eq):
+                if eq.is_symbolic() and eq.name() in all_states and p.match(eq.name()):
+                    return eq, 0.0
+
+                if eq.is_op(ca.OP_IF_ELSE_ZERO):
+                    cond = eq.dep(0)
+                    nested_eq = eq.dep(1)
+
+                    variable, value = extract_assignment(nested_eq)
+                    if variable is not None:
+                        return variable, ca.if_else(cond, value, 0)
 
                 if eq.n_dep() == 2 and (eq.is_op(ca.OP_SUB) or eq.is_op(ca.OP_ADD)):
+                    # Prefer eliminating alg states
                     if eq.dep(0).is_symbolic() and eq.dep(0).name() in alg_states and p.match(eq.dep(0).name()):
                         variable = eq.dep(0)
                         value = eq.dep(1)
                     elif eq.dep(1).is_symbolic() and eq.dep(1).name() in alg_states and p.match(eq.dep(1).name()):
                         variable = eq.dep(1)
                         value = eq.dep(0)
+                    elif eq.dep(0).is_symbolic() and eq.dep(0).name() in states and p.match(eq.dep(0).name()):
+                        variable = eq.dep(0)
+                        value = eq.dep(1)
+                    elif eq.dep(1).is_symbolic() and eq.dep(1).name() in states and p.match(eq.dep(1).name()):
+                        variable = eq.dep(1)
+                        value = eq.dep(0)
                     else:
                         variable = None
                         value = None
 
                     if variable is not None:
+                        if eq.is_op(ca.OP_SUB):
+                            return variable, value
+                        else:
+                            return variable, -value
+
+                if eq.n_dep() == 2 \
+                    and eq.is_op(ca.OP_ADD) \
+                    and eq.dep(0).is_op(ca.OP_IF_ELSE_ZERO) \
+                    and eq.dep(1).is_op(ca.OP_IF_ELSE_ZERO):
+                    # Once passed through _expand_simplify_mx, a CasADi if_else
+                    # is converted to a sum of two if_else_zero operations.
+
+                    cond_1 = eq.dep(0).dep(0)
+                    cond_2 = eq.dep(1).dep(0)
+                    nested_eq_1 = eq.dep(0).dep(1)
+                    nested_eq_2 = eq.dep(1).dep(1)
+
+                    variable_1, value_1 = extract_assignment(nested_eq_1)
+                    variable_2, value_2 = extract_assignment(nested_eq_2)
+
+                    if variable_1 is not None and variable_2 is not None and variable_1.name() == variable_2.name():
+                        return variable_1, ca.if_else(cond_1, value_1, 0) + ca.if_else(cond_2, value_2, 0)
+
+                # Nothing found
+                return None, None
+
+            variables = []
+            values = []
+
+            def get_derivative(expr):
+                if expr.is_constant():
+                    return 0.0
+                elif expr.is_symbolic():
+                    if expr.name() in states:
+                        return der_states[expr.name()].symbol
+                    elif expr.name() in alg_states:
+                        # This algebraic state must now become a differentiated state.
+                        states[expr.name()] = alg_states.pop(expr.name())
+                        der_sym = ca.MX.sym('der({})'.format(expr.name()))
+                        der_states[expr.name()] = Variable(der_sym, float)
+                        return der_sym
+                    else:
+                        return 0.0
+                else:
+                    # Differentiate using CasADi and chain rule
+                    orig_deps = ca.symvar(expr)
+                    deps = ca.vertcat(*orig_deps)
+                    J = ca.Function('J', [deps], [ca.jacobian(expr, deps)])
+                    J_sparsity = J.sparsity_out(0)
+                    der_deps = [get_derivative(dep) if J_sparsity.has_nz(0, j) else ca.DM.zeros(dep.size()) for j, dep in enumerate(orig_deps)]
+                    return ca.mtimes(J(deps), ca.vertcat(*der_deps))
+
+            reduced_equations = []
+            for eq in self.equations:
+                variable, value = extract_assignment(eq)
+                if variable is not None:
+                    if variable.name() in states:
+                        # Mark derivative state for replacement too.
+                        derivative = get_derivative(value)
+
+                        del states[variable.name()]
+
+                        variables.append(der_states.pop(variable.name()).symbol)
+                        values.append(derivative)
+
+                    elif variable.name() in alg_states:
                         del alg_states[variable.name()]
 
-                        variables.append(variable)
-                        if eq.is_op(ca.OP_SUB):
-                            values.append(value)
-                        else:
-                            values.append(-value)
+                    variables.append(variable)
+                    values.append(value)
 
-                        # Skip this equation
-                        continue
+                    # Skip this equation
+                    continue
 
                 # Keep this equation
                 reduced_equations.append(eq)
 
             # Eliminate alias variables
+            self.states = list(states.values())
             self.alg_states = list(alg_states.values())
+            self.der_states = list(der_states.values())
+
             self.equations = reduced_equations
 
-            if len(self.equations) > 0:
-                self.equations = ca.substitute(self.equations, variables, values)
-            if len(self.initial_equations) > 0:
-                self.initial_equations = ca.substitute(self.initial_equations, variables, values)
-            if len(self.delay_arguments) > 0:
-                self.delay_arguments = self._substitute_delay_arguments(self.delay_arguments, variables, values)
+            if len(values) > 0:
+                # Resolve expressions that include other, non-simple parameter
+                # expressions.
+                for i in range(SUBSTITUTE_LOOP_LIMIT):
+                    new_values = ca.substitute(values, variables, values)
+                    converged = ca.is_equal(ca.veccat(*values), ca.veccat(*new_values), CASADI_COMPARISON_DEPTH)
+                    values = new_values
+                    if converged:
+                        break
+                else:
+                    logger.warning("Substitution of expressions exceeded maximum iteration limit.")
 
-        if options.get('expand_vectors', False) and not options.get('expand_mx', False):
+                if len(self.equations) > 0:
+                    self.equations = ca.substitute(self.equations, variables, values)
+                if len(self.initial_equations) > 0:
+                    self.initial_equations = ca.substitute(self.initial_equations, variables, values)
+                if len(self.delay_arguments) > 0:
+                    self.delay_arguments = self._substitute_delay_arguments(self.delay_arguments, variables, values)
+
+        if options['expand_vectors'] and not options['expand_mx']:
             # If we are _not_ expanding MX to SX, we do the expansion of vectors here
             self._expand_vectors()
 
-        if options.get('factor_and_simplify_equations', False):
+        if options['factor_and_simplify_equations']:
             # Operations that preserve the equivalence of an equation
             # TODO: There may be more, but this is the most frequent set
             unary_ops = ca.OP_NEG, ca.OP_FABS, ca.OP_SQRT
@@ -588,7 +759,7 @@ class Model:
             # Store changes
             self.equations = simplified_equations
 
-        if options.get('detect_aliases', False):
+        if options['detect_aliases']:
             logger.info("Detecting aliases")
 
             states = OrderedDict([(s.symbol.name(), s) for s in self.states])
@@ -596,6 +767,7 @@ class Model:
             alg_states = OrderedDict([(s.symbol.name(), s) for s in self.alg_states])
             inputs = OrderedDict([(s.symbol.name(), s) for s in self.inputs])
             parameters = OrderedDict([(s.symbol.name(), s) for s in self.parameters])
+            constants = OrderedDict([(s.symbol.name(), s) for s in self.constants])
 
             all_states = OrderedDict()
             all_states.update(states)
@@ -603,9 +775,17 @@ class Model:
             all_states.update(alg_states)
             all_states.update(inputs)
             all_states.update(parameters)
+            all_states.update(constants)
 
             # For now, we only eliminate algebraic states.
-            do_not_eliminate = set(list(der_states) + list(states) + list(inputs) + list(parameters))
+            do_not_eliminate = set(list(der_states) + list(states) + list(inputs)
+                                   + list(parameters) + list(constants))
+
+            # When doing a `detect_aliases` pass multiple times, we have to avoid
+            # handling aliases we already handled before as their states no longer
+            # exist. To know which alias is new and which is old, we make a copy here
+            # of the alias relation.
+            old_alias_relation = self.alias_relation.copy()
 
             def _make_alias(deps, negative_alias=False):
                 """
@@ -662,7 +842,7 @@ class Model:
                         deps = ca.symvar(eq)
                         assert len(deps) == 2
                         if _make_alias(deps, eq.is_op(ca.OP_ADD)): continue
-                elif options.get('expand_vectors', False) and not options.get('expand_mx', False):
+                elif options['expand_vectors'] and not options['expand_mx']:
                     # Equation might have many "shadow" dependencies due to
                     # vector/array expansion. By using .expand() and SX
                     # symbols for the evaluation, we can figure out what the
@@ -708,6 +888,11 @@ class Model:
                 fixed = canonical_state.fixed
 
                 for alias in aliases:
+                    if len(old_alias_relation.aliases(alias)) > 1 and \
+                            alias not in old_alias_relation.canonical_variables:
+                        # We already handled this alias in a previous pass of `detect_aliases`
+                        continue
+
                     if alias[0] == '-':
                         sign = -1
                         alias = alias[1:]
@@ -726,9 +911,26 @@ class Model:
 
                     # If any of the aliases has a nondefault start value, apply it to
                     # the canonical state as well
-                    alias_state_start = ca.MX(alias_state.start)
-                    if alias_state_start.is_regular() and not alias_state_start.is_zero():
-                        start = sign * alias_state.start
+                    if not isinstance(alias_state.start, _DefaultValue):
+                        alias_start_mx = ca.MX(alias_state.start)
+                        if not isinstance(start, _DefaultValue):
+                            start_mx = ca.MX(start)
+                            # If the state already has a non-default start
+                            # attribute we check for conflicts.
+                            if (start_mx.is_constant() != ca.MX(alias_start_mx).is_constant()
+                                or (start_mx.is_symbolic() and str(start_mx) != str(sign * alias_start_mx))
+                                or start != alias_start_mx):
+
+                                logger.warning(
+                                    "Current start attribute of canonical variable '{}' ({})"
+                                    " conflicts with that of its alias '{}' ({})."
+                                    " Will keep existing value of {}."
+                                    .format(canonical, start, alias, alias_start_mx, start))
+                            else:
+                                # Alias has equal start attribute, so nothing to do
+                                pass
+                        else:
+                            start = sign * alias_state.start
 
                     # The intersection of all bound ranges applies
                     m = ca.fmax(m, alias_state.min if sign == 1 else -alias_state.max)
@@ -765,7 +967,7 @@ class Model:
             if len(self.delay_arguments) > 0:
                 self.delay_arguments = self._substitute_delay_arguments(self.delay_arguments, variables, values)
 
-        if options.get('reduce_affine_expression', False):
+        if options['reduce_affine_expression']:
             logger.info("Collapsing model into an affine expression")
 
             for equation_list in ['equations', 'initial_equations']:
@@ -803,7 +1005,7 @@ class Model:
                     equations = [ca.reshape(ca.mtimes(A, states_vector), equations.shape) + b]
                     setattr(self, equation_list, equations)
 
-        if options.get('expand_mx', False):
+        if options['expand_mx']:
             logger.info("Expanded MX functions will be returned")
             self._expand_mx_func = lambda x: x.expand()
 
@@ -843,7 +1045,13 @@ class Model:
             attribute_lists = [[] for i in range(len(CASADI_ATTRIBUTES))]
             for variable in variable_list:
                 for attribute_list_index, attribute in enumerate(CASADI_ATTRIBUTES):
-                    value = ca.MX(getattr(variable, attribute))
+                    value = getattr(variable, attribute)
+                    # Try casting to DM first, in case we have a nested list that needs to be interpreted as a matrix.
+                    try:
+                        value = ca.DM(value)
+                    except:
+                        pass
+                    value = ca.MX(value)
                     if value.is_zero():
                         value = zero
                     elif value.is_one():
@@ -919,3 +1127,4 @@ class Model:
                  ca.veccat(*self._symbols(self.constants)),
                  ca.veccat(*self._symbols(self.parameters))],
                 out_arguments))
+
